// ----------------------------------------------------------------------------
// Project:	Implementation Alternatives for Bottom-Up Evaluation
// Filename:	test8push.cpp
// Purpose:	Time Measurements for Rule Implementations: Push Method
// Last Change:	27.02.2010
// Language:	C++
// EMail:	brass@acm.org
// WWW:		http://www.informatik.uni-halle.de/~brass/
// Address:	Univ. Halle, Von-Seckendorff-Platz 1, D-06120 Halle, GERMANY
// Copyright:	(c) 2010 by Stefan Brass
//		Downloads and modifications for tests and scientific purposes
//		permitted, provided that some acknowledgement of my work
//		remains, and modified versions are clearly marked as such.
//		There is no warranty at all - this code may contain bugs.
// ----------------------------------------------------------------------------

// Example Program 8:
// p(X, Y) :- q(X, Y).
// p(X, Z) :- q(X, Y), p(Y, Z).
//
// q := {(i,i+1)|i = 1,...,1000}


#include <iostream>
using namespace std;

#if defined(_MSC_VER) || defined(_WINDOWS_)
#define WINDOWS
#else
#define UNIX
#endif

// ----------------------------------------------------------------------------
// Sizes for test:
// ----------------------------------------------------------------------------

const int Q_TUPLES =  1000;

const bool PRINT_EXTENSION = false;

const int STACKSIZE = 50;

// ----------------------------------------------------------------------------
// Timing Functions (UNIX Version):
// ----------------------------------------------------------------------------

#ifdef UNIX
#include <sys/time.h>
#include <time.h>

class timer {
public:
	void start()
	{
		gettimeofday(&Time_Start, 0);
		Clock_Start = clock();
	}

	void end()
	{
		gettimeofday(&Time_End, 0);
		Clock_End = clock();
	}
	
	long int realtime_msec()
	{
		long sec = Time_End.tv_sec - Time_Start.tv_sec;
		long usec = Time_End.tv_usec - Time_Start.tv_usec;
		if(usec < 0) {
			usec += 1000000;
			sec -= 1;
		}
		return sec * 1000 + (usec + 500) / 1000;
	}
	long int cputime_msec()
	{
		long ticks = Clock_End - Clock_Start;
		return ticks / (CLOCKS_PER_SEC / 1000);
	}

private:
	timeval Time_Start, Time_End;
	clock_t Clock_Start, Clock_End;
};
#endif

// ----------------------------------------------------------------------------
// Timing Functions (Windows Version):
// ----------------------------------------------------------------------------

#ifdef WINDOWS
#include <windows.h>
#include <time.h>

class timer {
public:
	void start()
	{
		GetSystemTimeAsFileTime(&(Start.ft));
		Clock_Start = clock();
	}

	void end()
	{
		GetSystemTimeAsFileTime(&(End.ft));
		Clock_End = clock();
	}
	
	long int realtime_msec()
	{
		return static_cast<long> ((End.ns100 - Start.ns100) / 10000);
	}
	long int cputime_msec()
	{
		long ticks = Clock_End - Clock_Start;
		return ticks / (CLOCKS_PER_SEC / 1000);
	}

private:
	union {
		long long ns100;
		FILETIME ft;
	} Start, End;
	clock_t Clock_Start, Clock_End;
};
#endif


// ----------------------------------------------------------------------------
// Stack (of int):
// ----------------------------------------------------------------------------


class stack {

private:
	static int stack_arr[STACKSIZE];
	static int *stack_p;
	static int stack_free;

public:
	static inline void stack_init()
	{
		stack_p = stack_arr;
		stack_free = STACKSIZE;
	}

	static inline void push(int i)
	{
		if(stack_free > 0) {
			stack_free--;
			*(stack_p++) = i;
		}
		else {
			cout << "Stack overflow\n";
			exit(1);
		}
	}

	static inline int pop()
	{
		if(stack_p > stack_arr) {
			stack_free++;
			return *(--stack_p);
		}
		else {
			cout << "Stack underflow\n";
			exit(1);
		}
	}

	static inline bool is_empty()
	{
		return (stack_p == stack_arr);
	}
};

int stack::stack_arr[STACKSIZE];
int *stack::stack_p;
int stack::stack_free;


//-----------------------------------------------------------------------------
// Predicate q: Tuples generated by computation
//-----------------------------------------------------------------------------

class q_cursor {
public:
	q_cursor() { }

	void open() { i = 0; }

	bool fetch()
	{
		if(i < Q_TUPLES) {
			i++;
			return true;
		}
		else
			return false;
	}

	int col_1() { return i; }

	int col_2() { return i+1; }

	void close() {}

	void push() { stack::push(i); }

	void pop() { i = stack::pop(); }

private:
	int i;
};


// ----------------------------------------------------------------------------
// Predicate p - Push Method
// ----------------------------------------------------------------------------


class p_cursor {
private:

	// Continuation labels:
	typedef enum {CONT_init, CONT_r1_q, CONT_p, CONT_r2, CONT_r2_q }
		cont_t;

	// Functions for using the global stack:
	inline void push_int(int i)
	{
		stack::push(i);
	}

	inline int pop_int()
	{
		return stack::pop();
	}

	inline void push_cont(cont_t t)
	{
		stack::push(t);
	}

	inline cont_t pop_cont()
	{
		return (cont_t) stack::pop();
	}

public:
	void open()
	{
		stack::stack_init();
		push_cont(CONT_init);
	}

	bool fetch()
	{
		int Y;

	backtrack:
		if(stack::is_empty())
			return false;
		switch(pop_cont()) {
			case CONT_init: goto cont_init;
			case CONT_r1_q: goto cont_r1_q;
			case CONT_p: goto cont_p;
			case CONT_r2: goto cont_r2;
			case CONT_r2_q: goto cont_r2_q;
			default: cout << "Impossible cont\n"; return false;
		}

	cont_init:
		// p(X, Y) :- q(X, Y).
		q_c.open();
		if(q_c.fetch()) {
			p_1 = q_c.col_1();
			p_2 = q_c.col_2();
			if(q_c.fetch()) {
				q_c.push();
				push_cont(CONT_r1_q);
			}
			goto cont_p;
		}
		return false; // Last initialization rule

	cont_r1_q:
		q_c.pop();
		p_1 = q_c.col_1();
		p_2 = q_c.col_2();
		if(q_c.fetch()) {
			q_c.push();
			push_cont(CONT_r1_q);
		}
		goto cont_p;

	cont_p:
		push_cont(CONT_r2);
		return true;
	
	cont_r2:
		// p(X, Z) :- q(X, Y), p(Y, Z).
		q_c.open();
		while(q_c.fetch()) {
			if(q_c.col_2() == p_1) { // Join over Y
				Y = p_1;
				p_1 = q_c.col_1();
				// p_2 = p_2;
				while(q_c.fetch()) {
					if(q_c.col_2() == Y) {
						q_c.push();
						push_int(Y);
						push_int(p_2);
						push_cont(CONT_r2_q);
						break;
					}
				}
				goto cont_p;
			}
		}
		goto backtrack;

	cont_r2_q:
		q_c.pop();
		p_2 = pop_int();
		Y = pop_int();
		p_1 = q_c.col_1();
		// p_2 = p_2;
		while(q_c.fetch()) {
			if(q_c.col_2() == Y) {
				q_c.push();
				push_int(Y);
				push_int(p_2);
				push_cont(CONT_r2_q);
				break;
			}
		}
		goto cont_p;
	}

    int col_1() { return p_1; }
    int col_2() { return p_2; }

    void close()
    {
    }
private:
	// Arguments:
	int p_1, p_2;

	// Cursors:
	q_cursor q_c;
};


//-----------------------------------------------------------------------------
// Main Program:
//-----------------------------------------------------------------------------

int main()
{
	timer t;
	int c1 = 0;
	int c2 = 0;
	int num_tuples = 0;

	cout << "Test of Method 3: Push [Program 8]\n";
	t.start();
	stack::stack_init();
	p_cursor p_c;
	p_c.open();
	while(p_c.fetch()) {
		c1 += p_c.col_1();
		c2 += p_c.col_2();
		num_tuples++;
	}
	p_c.close();
	t.end();

	if(PRINT_EXTENSION) {
		cout << "Extension of p:\n\n";
		p_c.open();
		while(p_c.fetch()) {
			cout << "p(" << p_c.col_1() << ", " << p_c.col_2()
				<< ").\n";
		}
		p_c.close();
	}

	cout << "\n\nTotal number of tuples: " << num_tuples << "\n";
	cout << "Column sums: col1=" << c1 << ", col2=" << c2 << "\n";
	cout << "\nUsed time: " << t.realtime_msec() << " ms,  "
		<< "CPU: " << t.cputime_msec() << " ms.\n\n\n";

	return 0;
}


